import os
import glob
import shutil
import unittest

import matplotlib.pyplot as plt

from adamspy.postprocess.ppt import manually_remove_spikes, manually_remove_spikes_batch

TEST_ANALYSIS_DIR = os.path.join('test', 'files', 'analysis_with_spikes')

class Test_ManuallyRemoveSpikes(unittest.TestCase):

    def setUp(self):
        self.temp_ans_dir = TEST_ANALYSIS_DIR + '_tmp'        
        self._setup_files()
        self.res_to_edit =os.path.join(self.temp_ans_dir, 'test_analysis_with_spikes.res')

    def test_spike_removed_from_results(self):
        results_to_clean = {'torque': ['Q']}
        results = manually_remove_spikes(self.res_to_edit, results_to_clean)
                
        self.assertLess(max(results['torque']['Q']), .01)

    def test_spike_removed_from_other_result_comps(self):
        results_to_clean = {'torque': ['Q'], 'torque_spike': ['FX', 'FY', 'FZ', 'TX', 'TY', 'TZ']}
        results = manually_remove_spikes(self.res_to_edit, results_to_clean)
        
        self.assertTrue(all([max(results['torque']['Q']) < .01, max(results['torque_spike']['TZ']) < .01]))

    def test_spike_removed_from_other_result_comps_without_checking(self):
        results_to_clean = {'torque': ['Q'], 'torque_spike': ['FX', 'FY', 'FZ', 'TX', 'TY', 'TZ']}
        results_to_check = {'torque_spike': ['FX', 'FY', 'FZ', 'TX', 'TY', 'TZ']}
        results = manually_remove_spikes(self.res_to_edit, results_to_clean, reqs_to_check=results_to_check)
        
        self.assertTrue(all([max(results['torque']['Q']) < .01, max(results['torque_spike']['TZ']) < .01]))

    def test_spike_removed_from_results_batch(self):
        results_to_clean = {
            'torque':       ['Q'], 
            'torque_spike': ['TX', 'TY', 'TZ'], 
            'PART_2_XFORM': ['X', 'Y', 'Z', 'VX', 'VY', 'VZ', 'ACCX', 'ACCY', 'ACCZ'], 
            'PART_3_XFORM': ['X', 'Y', 'Z', 'VX', 'VY', 'VZ', 'ACCX', 'ACCY', 'ACCZ']
            }

        results_to_check = [
            {
                # 'torque': ['Q'],
                'torque_spike': ['TX', 'TY', 'TZ']
            },
            {
                'PART_2_XFORM': ['X', 'Y', 'Z'],
                'PART_3_XFORM': ['X', 'Y', 'Z']
            },
            {
                'PART_2_XFORM': ['VX', 'VY', 'VZ'],
                'PART_3_XFORM': ['VX', 'VY', 'VZ']
            },
            {
                'PART_2_XFORM': ['ACCX', 'ACCY', 'ACCZ'],
                'PART_3_XFORM': ['ACCX', 'ACCY', 'ACCZ']
            }
        ]

        results = manually_remove_spikes_batch(self.res_to_edit, results_to_clean, results_to_check)
                
        self.assertLess(max(results['torque']['Q']), .01)
        
    def tearDown(self):
        """Deletes the temporary working directory.

        """
        shutil.rmtree(self.temp_ans_dir)
            
    def _setup_files(self):
        """Copies the test analysis files to a temporary working directory

        """
        if os.path.exists(self.temp_ans_dir) is True:
            shutil.rmtree(self.temp_ans_dir)

        os.makedirs(self.temp_ans_dir)
        
        for file in glob.glob(os.path.join(TEST_ANALYSIS_DIR, '*')):
            shutil.copyfile(file, os.path.join(self.temp_ans_dir, os.path.split(file)[-1]))   


"""
"""
import os
import glob
from pathlib import Path
from tempfile import TemporaryDirectory
import time
import shutil
import unittest

from adamspy.adamspy import set_n_threads, get_n_threads, solve, XMT_PATTERN, modify_xmt_path, AviewError
from adamspy.aviewpy import run_commands, run_script, START_SCRIPT_NAMES

from test import TEST_ADM_FILE, EXISTING_MODEL_DIR, TEST_CMD_XMT


TEST_ADM_FILE_TO_CHANGE = os.path.join(os.path.split(TEST_ADM_FILE)[0], 'test_modified.adm')


class Test_RunCommands(unittest.TestCase):

    def setUp(self):
        return

    def test_run_commands_export(self):

        cmds = [
            'model create model=model_1',
            'marker create marker=.model_1.ground.marker_1',
            'file command write ent=.model_1 file="model_1.cmd"'
        ]

        with TemporaryDirectory() as twd_:
            twd = Path(twd_).resolve()
            run_commands(cmds, twd)

            self.assertTrue((twd / 'model_1.cmd').exists())

    def test_run_commands_raises_error(self):

        cmds = [
            'model create model=model_1',
            'marker create marker=.model_1.marker_1',
        ]
        with self.assertRaises(AviewError), TemporaryDirectory() as twd_:
            twd = Path(twd_).resolve()
            run_commands(cmds, twd)

    def test_run_commands_ignores_startup(self):
        
        cmds = [
            'model create model=model_1',
            'marker create marker=.model_1.ground.marker_1',
        ]

        with TemporaryDirectory() as twd_:
            twd = Path(twd_).resolve()

            for file in START_SCRIPT_NAMES:
                (twd / file).write_text(
                    f'var set var=tmp str=(eval(str_print("! ERROR: {file} not ignored!")))'
                )
            run_commands(cmds, twd)

    def tearDown(self):
        return


class Test_RunScript(unittest.TestCase):

    def setUp(self):
        return

    def test_run_script_export(self):

        cmds = [
            'model create model=model_1',
            'marker create marker=.model_1.ground.marker_1',
            'file command write ent=.model_1 file="model_1.cmd"'
        ]

        with TemporaryDirectory() as twd_:
            twd = Path(twd_).resolve()

            # Write the commands to a script
            (twd / 'script.cmd').write_text('\n'.join(cmds))
            
            # Run the script
            run_script((twd / 'script.cmd'), twd)
            
            # Ensure the model exists
            self.assertTrue((twd / 'model_1.cmd').exists())

    def test_run_script_cwd_implied(self):

        cmds = [
            'model create model=model_1',
            'marker create marker=.model_1.ground.marker_1',
            'file command write ent=.model_1 file="model_1.cmd"'
        ]

        with TemporaryDirectory() as twd_:
            twd = Path(twd_).resolve()

            # Write the commands to a script
            (twd / 'script.cmd').write_text('\n'.join(cmds))
            
            # Run the script
            run_script((twd / 'script.cmd'))
            
            # Ensure the model exists
            self.assertTrue((twd / 'model_1.cmd').exists())

    def test_run_script_raises_error(self):

        cmds = [
            'model create model=model_1',
            'marker create marker=.model_1.marker_1',
        ]
        with self.assertRaises(AviewError), TemporaryDirectory() as twd_:
            twd = Path(twd_).resolve()

            # Write the commands to a script
            (twd / 'script.cmd').write_text('\n'.join(cmds))

            run_script((twd / 'script.cmd'), twd)

    def test_run_script_preserves_log(self):

        cmds = [
            'model create model=model_1',
            'marker create marker=.model_1.ground.marker_1',
        ]
        with TemporaryDirectory() as twd_:
            twd = Path(twd_).resolve()

            # Write the commands to a script
            (twd / 'script.cmd').write_text('\n'.join(cmds))

            # Create a dummy log
            log_text = 'MUST PRESERVE THIS FILE'
            (twd / 'aview.log').write_text(log_text)

            run_script((twd / 'script.cmd'), twd)

            self.assertEqual((twd / 'aview.log').read_text(), log_text)

    def test_run_script_ignores_start_scripts(self):

        cmds = [
            'model create model=model_1',
            'marker create marker=.model_1.ground.marker_1',
        ]
        with TemporaryDirectory() as twd_:
            twd = Path(twd_).resolve()

            # Write the commands to a script
            (twd / 'script.cmd').write_text('\n'.join(cmds))

            for file in START_SCRIPT_NAMES:
                (twd / file).write_text(
                    f'var set var=tmp str=(eval(str_print("! ERROR: {file} not ignored!")))'
                )

            run_script((twd / 'script.cmd'), twd)

            self.assertListEqual(
                [(twd / f).read_text() for f in START_SCRIPT_NAMES],
                [f'var set var=tmp str=(eval(str_print("! ERROR: {f} not ignored!")))'
                 for f in START_SCRIPT_NAMES]
            )

    def tearDown(self):
        return


class Test_SetNThreads(unittest.TestCase):
    """
    """
    def setUp(self):
        # Delete the modified adm file if it exists
        try:
            os.remove(TEST_ADM_FILE_TO_CHANGE)
        except FileNotFoundError:
            pass

        # Copy the adm file
        shutil.copyfile(TEST_ADM_FILE, TEST_ADM_FILE_TO_CHANGE)

    def test_set_n_threads(self):
        """Tests that set_n_threads correctly changes the adm file.

        """
        set_n_threads(TEST_ADM_FILE_TO_CHANGE, 10)
        with open(TEST_ADM_FILE_TO_CHANGE, 'r') as fid:
            lines = fid.readlines()

        threads_line = lines[12336].strip()

        self.assertEqual(threads_line, ', NTHREADS = 10')

    def test_get_n_threads(self):
        """Tests that get_n_threads correctly gets the number of threads defined in an adm file.

        """
        n_threads = get_n_threads(TEST_ADM_FILE)
        self.assertEqual(n_threads, 4)

    def tearDown(self):
        # Delete the modified adm file if it exists
        try:
            os.remove(TEST_ADM_FILE_TO_CHANGE)
        except FileNotFoundError:
            pass


class Test_Solve(unittest.TestCase):
    """
    """
    def setUp(self):
        for file in glob.glob(os.path.join(EXISTING_MODEL_DIR, '*')):
            ext = os.path.splitext(file)[-1]
            if ext not in ['.acf', '.adm', '.xmt_txt']:
                try:
                    os.remove(file)
                except FileNotFoundError:
                    pass
                except PermissionError:
                    pass

    def test_solve_wait(self):
        """Tests that adamspy.adamspy.solve() works when wait=True

        """
        acf_file = glob.glob(os.path.join(EXISTING_MODEL_DIR, '*.acf'))[0]
        solve(acf_file, wait=True)
        res_files = glob.glob(os.path.join(EXISTING_MODEL_DIR, '*.res'))
        self.assertTrue(res_files != [])

    def test_solve_dont_wait(self):
        """Tests that adamspy.adamspy.solve() works when wait=False

        """
        acf_file = glob.glob(os.path.join(EXISTING_MODEL_DIR, '*.acf'))[0]
        solve(acf_file, wait=False)

        # Check that the res file isn't there yet
        res_files = glob.glob(os.path.join(EXISTING_MODEL_DIR, '*.res'))
        if res_files != []:
            self.fail('Solve did not wait!')

        # Wait 60 seconds for the res file to show up
        for _i in range(60):
            res_files = glob.glob(os.path.join(EXISTING_MODEL_DIR, '*.res'))
            if res_files != []:
                break
            time.sleep(1)

        self.assertTrue(res_files != [])

    def tearDown(self):
        for file in glob.glob(os.path.join(EXISTING_MODEL_DIR, '*')):
            ext = os.path.splitext(file)[-1]
            if ext not in ['.acf', '.adm', '.xmt_txt']:
                try:
                    os.remove(file)
                except FileNotFoundError:
                    pass
                except PermissionError:
                    pass


class Test_CMD(unittest.TestCase):
    """
    """
    def setUp(self):
        return

    def test_xmt_pattern(self):
        """
        """
        xmt_line = '   file_name = "FRS3EQ_case1_LAUL.xmt_txt"  &'
        match = XMT_PATTERN.match(xmt_line)

        self.assertTrue(bool(match))

    def test_modify_xmt_path(self):
        """
        """
        path_to_change_to = os.path.join('path', 'to', 'bogus.xmt_txt')
        original_path = modify_xmt_path(TEST_CMD_XMT, path_to_change_to)
        new_path = modify_xmt_path(TEST_CMD_XMT, original_path)

        self.assertListEqual([original_path, new_path], ['FRS3EQ_case1_LAUL.xmt_txt', path_to_change_to])

    def tearDown(self):
        return



from __future__ import annotations
import re
from io import StringIO
from typing import Dict, Tuple

from numpy import ceil, sqrt, linalg
import pandas as pd


MODE_LABEL = re.compile('^\\s*Energy table for MODE\\s*\\d+$', flags=re.MULTILINE)
MODAL_SCALAR_LABELS = {
    'zeta': re.compile('^\\s*Damping ratio\\s*=.*$', flags=re.MULTILINE),
    'w_n': re.compile('^\\s*Undamped natural freq\\.\\s*=.*$', flags=re.MULTILINE),
    'K': re.compile('^\\s*Generalized stiffness\\s*=.*$', flags=re.MULTILINE),
    'E': re.compile('^\\s*Kinetic energy\\s*=.*$', flags=re.MULTILINE)
}

MODAL_TABLE_LABELS = {
    'coords': re.compile('^\\s*Normalized coordinates.*$', flags=re.MULTILINE),
    'KE': re.compile('^\\s*Percentage distribution of Kinetic energy.*$', flags=re.MULTILINE),
    'SE': re.compile('^\\s*Percentage distribution of strain energy.*$', flags=re.MULTILINE),
    'DE': re.compile('^\\s*Percentage distribution of dissipative energy.*$', flags=re.MULTILINE)
}

# MODE_TABLE_HORIZ_SEP = re.compile('^\\s*\\+-+\\+\\s*$', flags=re.MULTILINE)
MODE_TABLE_HORIZ_SEP = re.compile('^\\s*[\\+-]+\\s*$', flags=re.MULTILINE)

MODE_TABLE_NAME_MAP = {
    **{f'{d}_mag.': f'{d}_magnitude' for d in ['X', 'Y', 'Z', 'RX', 'RY', 'RZ']},
    **{f'{d}_ph.': f'{d}_phase' for d in ['X', 'Y', 'Z', 'RX', 'RY', 'RZ']},
}


class Mode():
    def __init__(self, scalars: dict, tables: dict):
        self.zeta = scalars['zeta']
        self.w_n = scalars['w_n']
        self.K = scalars['K']
        self.E = scalars['E']
        self.coords: pd.DataFrame = tables['coords'] if 'coords' in tables else None
        self.kinetic_energy: pd.DataFrame = tables['KE'] if 'KE' in tables else None
        self.strain_energy: pd.DataFrame = tables['SE'] if 'SE' in tables else None
        self.dissipative_energy: pd.DataFrame = tables['DE'] if 'DE' in tables else None

        if self.coords is not None:
            self._add_mag_to_coords()
    
    def __repr__(self) -> str:

        text = ' | '.join([
            super(Mode, self).__repr__(),
            f'w_n = {self.w_n}',
            f'K = {self.K}',
            f'E = {self.E}',
            f'zeta = {self.zeta}',
        ])

        return text

    def _add_mag_to_coords(self):
        trans = pd.DataFrame(self.coords[['X_magnitude', 'Y_magnitude', 'Z_magnitude']]).dropna()
        M_magnitude = sqrt(trans['X_magnitude']**2
                           + trans['Y_magnitude']**2
                           + trans['Z_magnitude']**2).rename('M_magnitude')

        rot = pd.DataFrame(self.coords[['RX_magnitude', 'RY_magnitude', 'RZ_magnitude']]).dropna()
        RM_magnitude = sqrt(rot['RX_magnitude']**2
                            + rot['RY_magnitude']**2
                            + rot['RZ_magnitude']**2).rename('RM_Magnitude')

        self.coords = (self.coords
                       .merge(M_magnitude, left_index=True, right_index=True, how='left')
                       .merge(RM_magnitude, left_index=True, right_index=True, how='left'))

    def compare_coords(self, other_mode: Mode):
        err = pd.DataFrame(self.coords.set_index('solver_name')[[f'{d}_magnitude' for d in ('X', 'Y', 'Z')]]
                           - other_mode.coords.set_index('solver_name')[[f'{d}_magnitude' for d in ('X', 'Y', 'Z')]])

        err = (err
               .apply(linalg.norm, axis=1)
               .sum())

        return err


class Modal():

    def __init__(self, modes: Dict[int, Mode]):
        self.modes = modes

        self.w_n = {i: m.w_n for i, m in self.modes.items()}
        self.E = {i: m.E for i, m in self.modes.items()}
        self.K = {i: m.K for i, m in self.modes.items()}
        self.zeta = {i: m.zeta for i, m in self.modes.items()}
        self.scalars = pd.DataFrame({
            'w_n': self.w_n,
            'E': self.E,
            'K': self.K,
            'zeta': self.zeta,
        })

    def filter_eigs(self, wn_min, wn_max):
        for num in list(self.modes):
            if self.modes[num].w_n < wn_min or self.modes[num].w_n > wn_max:
                _ = self.modes.pop(num)

    def remove_zero_eigs(self, threshold=.5):
        for num in list(self.modes):
            if self.modes[num].w_n < threshold:
                _ = self.modes.pop(num)

    def remove_duplicate_eigs(self, tolerance=.001):
        unique = {}

        for num, mode in self.modes.items():
            if all([mode.w_n - w_n > tolerance for w_n in unique.values()]) or unique == {}:
                unique[num] = mode.w_n

        self.modes = {num: self.modes[num] for num in unique}

    def get_modes_where_parts_participate(self, part_suffix, table, column, threshold=.9, sum_or_max='max'):
        """Returns a dictionary where the keys are mode numbers and the values are `:obj:Mode` objects in which any part with the suffix `part_suffix` has at least `threshold` percent of the total.

        Parameters
        ----------
        part_suffix : str
            Suffix of parts to check for
        table : str
            Table type. Options are 'coords', 'kinetic_energy', 'strain_energy', 'dissipative_energy'
        column : str
            Column name from the table
        threshold : float
            Part's value must be greater than `threshold` multiplied the sum or max (depending on `sum_or_max` value) of the column in the table to be considered participating
        sum_or_max : str
            Must be 'sum' or 'max'

        Returns
        -------
        dict
            Dictionary where the keys are mode numbers and the values are `:obj:Mode` objects

        """
        modes_with_participation = {}

        for mode_number, mode in self.modes.items():

            # Sort the kinetic energy table by decending sum
            tbl_obj = getattr(mode, table)
            sorted_tbl = tbl_obj.sort_values(column, ascending=False).reset_index(drop=True)

            # Get the value to compare against
            value = sorted_tbl[column].max() if sum_or_max.lower() == 'max' else sorted_tbl[column].sum()

            filtered_tbl = sorted_tbl[sorted_tbl['view_name'].str.split('.', expand=True)[1].str.startswith(part_suffix)].reset_index(drop=True)

            if filtered_tbl[column][0] >= value*threshold:
                modes_with_participation[mode_number] = mode

        return modes_with_participation

    def find_pairs(self, other: Modal):
        pairs = {}
        errors = {}
        for num_a, mode_a in self.modes.items():
            self._find_pair(other.modes, mode_a, pairs, errors, num_a)

        return pairs

    def _find_pair(self, other_modes: Dict[int, Mode], mode_a, pairs, errors, num_a):
        errs = (pd.Series([mode_a.compare_coords(mode_b) for mode_b in other_modes.values()])
                .set_axis(list(other_modes)))

        for num_b, err in errs.sort_values().iteritems():
            
            if num_b in pairs.values():
                # If this mode is already paired
                current_num_a = [na for na, nb in pairs.items() if nb == num_b][0]
                current_err = errors[current_num_a]

                # If the new pairing is better
                if err < current_err:
                    pairs[num_a] = num_b
                    errors[num_a] = err
                    self._find_pair(other_modes, self.modes[current_num_a], pairs, errors, current_num_a)
                    return

            else:
                pairs[num_a] = num_b
                errors[num_a] = err
                return
        

def parse_table_file(filename):
    with open(filename, 'r') as fid:
        lines = fid.readlines()

    mode_lines_dict = _split_by_mode(lines)

    modes_dict = {}
    for mode_number, mode_lines in mode_lines_dict.items():
        scalars_dict, tables_dict = _parse_mode_lines(mode_lines)
        modes_dict[mode_number] = Mode(scalars_dict, tables_dict)

    return Modal(modes_dict)


def _split_by_mode(lines: list) -> dict:

    modes = []
    start_lines = []

    # Get the mode numbers and the start lines
    for i_line, line in enumerate(lines):
        if MODE_LABEL.match(line):
            mode_number = int(line.strip().replace('\n', '').split(' ')[-1])
            start_line = i_line - 1

            modes.append(mode_number)
            start_lines.append(start_line)

    # Create a dictionary of lists of lines for each mode
    mode_lines = {}
    for mode, start_line, stop_line in zip(modes, start_lines, [sl for sl in start_lines[1:]]+[-1]):
        mode_lines[mode] = lines[start_line:stop_line]

    return mode_lines


def _parse_mode_lines(mode_lines: list) -> Tuple[dict, Dict[str, pd.DataFrame]]:
    """Given a list of strings representing lines in an Adams/Linear modal output file for a single
    mode, returns dictionaries of the modal information.

    Parameters
    ----------
    mode_lines : list
        List of strings representing lines in an Adams/Linear modal output file for a single mode.

    Returns
    -------
    dict
        Dictinary of the scalars that describe the mode
    dict
        Dictionary of dataframes of the the lines for each mode table.

    """
    mode_scalars_dict = {}
    table_names = []
    tbl_st_ln_nums = []

    for i_line, line in enumerate(mode_lines):
        for scalar_name, pattern in MODAL_SCALAR_LABELS.items():
            if pattern.match(line):
                mode_scalars_dict[scalar_name] = float(line.split('=')[-1].strip().split(' ')[0])
                break

        for table_name, pattern in MODAL_TABLE_LABELS.items():
            if pattern.match(line):
                table_names.append(table_name)
                tbl_st_ln_nums.append(i_line)
                break

    # Create a dictionary of pandas dataframes
    tbl_dict = {}
    for tbl_name, st_ln_num, sp_ln_num in zip(table_names, tbl_st_ln_nums, [sl for sl in tbl_st_ln_nums[1:]]+[None]):

        # Get a pandas dataframe of the table
        tbl_dict[tbl_name] = _parse_table_lines(mode_lines[st_ln_num:sp_ln_num]) if sp_ln_num is not None else _parse_table_lines(mode_lines[st_ln_num:])

    return mode_scalars_dict, tbl_dict


def _parse_table_lines(table_lines: list):
    """Given a list of strings representing lines in an Adams/Linear modal output file for a single
    mode, returns a pandas table of the data


    Parameters
    ----------
    table_lines : list
        list of strings representing lines in an Adams/Linear modal output file for a single mode

    Returns
    -------
    pd.DataFrame
        Pandas `:obj:pd.DataFrame` representing the mode table

    """
    idx_horiz_sep = [i for i, line in enumerate(table_lines) if MODE_TABLE_HORIZ_SEP.match(line)][:3]

    # Get the header lines should be one or two lines
    header_lines = table_lines[idx_horiz_sep[0]+1:idx_horiz_sep[1]]

    # Get the column names
    column_names = _get_column_names(header_lines)
    matches = list(re.finditer('\\s\\w', header_lines[-1].replace('|', '')))
    colspecs = [(0, matches[0].end()-2)]
    for last_match, match in zip(matches[:-1], matches[1:]):
        colspecs += [(last_match.end()-2, match.end()-2)]

    # Prep lines for input to read csv by removing | and replaceing white space with commas
    prepped_lines = []
    for line in table_lines[idx_horiz_sep[1]+1:idx_horiz_sep[2]]:
        prepped_line = line.replace('|', '')
        # prepped_line = prepped_line.strip()
        # prepped_line = re.sub('(?<=|.*)  (?=|.*)', ',', prepped_line)
        # prepped_line = re.sub('\\s{1,11}', ',', prepped_line)
        # prepped_line = re.sub('(?<=\\d)-', ',-', prepped_line)
        prepped_lines.append(prepped_line)
    table_text = '\n'.join(prepped_lines)

    # Get the table as a dataframe
    return pd.read_fwf(StringIO(table_text), colspecs=colspecs, names=column_names).rename(columns=MODE_TABLE_NAME_MAP)


def _get_column_names(header_lines):
    if len(header_lines) == 2:
        # If the header has two lines
        first_headers_list = re.split('\\s+', header_lines[0].replace('|', '').strip())
        second_headers_list = re.split('\\s+', header_lines[1].replace('|', '').strip())

        column_names = ['solver_name'] + ['{}_{}'.format(first_headers_list[int(ceil((i+1)/2)-1)], sec_head) for i, sec_head in enumerate(second_headers_list)] + ['view_name']

    elif len(header_lines) == 1:
        # If the header has one line
        column_names = ['solver_name'] + re.split('\\s+', header_lines[0].replace('|', '').strip()) + ['view_name']

    return column_names



import os
import unittest


from adamspy.postprocess.msg import get_modes

TEST_MSG_FILE = os.path.join(os.getcwd(), 'test', 'files', 'get_modes', 'LC_4_int_modal.msg')

class Test_GetModes(unittest.TestCase):

    def setUp(self):
        self.modes_unsorted = get_modes(TEST_MSG_FILE, sort_by_wn=False, underdamped_only=False)
        self.modes_sorted = get_modes(TEST_MSG_FILE)

    def test_first_ten_reals(self):        
        first_ten_reals = self.modes_unsorted['Real'][:10]
        expected_reals = [
            4.32782E+00,
            7.56550E+00,
            -8.63392E+00,
            -1.20591E+01,
            -1.23835E+01,
            -1.23848E+01,
            -1.25394E+01,
            -1.26904E+01,
            -1.27002E+01,
            -1.28625E+01
        ]

        self.assertListEqual(first_ten_reals, expected_reals)

    def test_first_ten_freqs(self):        
        first_ten_reals = self.modes_unsorted['Wn'][:10]
        expected_reals = [4.32782, 7.5655, 8.63392, 12.0591, 12.3835, 12.3848, 12.5394, 12.6904, 12.7002, 12.8625]

        self.assertListEqual(first_ten_reals, expected_reals)

    def test_last_real(self):        
        last_real = self.modes_unsorted['Real'][-1]
        expected_real = -1.59054E+04

        self.assertEqual(last_real, expected_real)

    def test_first_ten_freqs_sorted(self):
        first_ten_reals = self.modes_sorted['Wn'][:10]
        expected_reals = [10.315066083249297, 13.185346399605889, 14.901002538138165, 15.890317234965904, 15.891802089691907, 15.900500613152028, 15.903637885528738, 15.904800161856416, 15.91000179855844, 15.910600035681854]

        self.assertListEqual(first_ten_reals, expected_reals)    

    def tearDown(self):
        return
        


import os
import csv

import numpy as np
import Adams                            # pylint: disable=import-error

def write_results(file,  reqs_to_get, output_file, t_min=None, t_max=None):                             #pylint: disable=redefined-outer-name
    """Opens Adams/View, loads results files, and writes the specified results to an output file.
    
    Parameters
    ----------
    file : list
        List of results or requests filenames
    reqs_to_get : dict
        Dictionary of results to export.  The keys should be results and the values should be result componenets
    t_min : float (Optional)
        Minimum time to export.  (first time in the dataset by default)
    t_max : float (Optional)
        Maximum time to export  (last time in the dataset by default)
    output_file : str
        Name of output file

    """
    # Initialize a dictionary for the results 
    res_dict = {}
        
    # Load Results    
    if os.path.exists(os.path.splitext(file)[0] + '.adm'):
        Adams.execute_cmd('file adams_data_set  read file_name="{}"'.format(os.path.splitext(file)[0] + '.adm'))
        Adams.execute_cmd('file analysis read file_name="{}" model_name={}'.format(file, os.path.splitext(os.path.split(file)[-1])[0]))
    else:
        Adams.execute_cmd('file analysis read file_name="{}"'.format(file))

    # Get the model object
    mod = Adams.Models.get([mn for mn in Adams.Models][0])
    
    # Get the analysis object
    ans = mod.Analyses.get([an for an in mod.Analyses][0])

    time = ans.results.get('TIME').values
    _time_np = np.asarray(time)
    i_min = 0 if t_min is None else np.argmax(_time_np >= t_min)
    i_max = len(time)-1 if t_max is None else np.argmax(_time_np >= t_max)

    # Store the time values
    res_dict['time'] = time[i_min:i_max]    

    # Loop over the requsted results
    for res_name, res_comps in reqs_to_get.items():
        print(f'res_comps={res_comps}')
        
        # Get the result object handle
        res = ans.results.get(res_name) if res_name in ans.results.keys() else get_req_from_comment(res_name, ans)

        for res_comp in res_comps:

            print(f'res_comp={res_comp}')
            print(f'req = {res}')
            
            values = res.get(res_comp).values[i_min:i_max]
            res_dict[f'{res_name}_{res_comp}'] = values
    
    # Write to a csv file
    with open(output_file, 'w', newline='') as outfile:
        writer = csv.writer(outfile)
        writer.writerow(res_dict.keys())
        writer.writerows(zip(*res_dict.values()))

def get_result_file_type(file):        
    """Checks if `file` is a results file or a request file.
    
    Parameters
    ----------
    file : str
        Results/request file name.
    
    Returns
    -------
    str
        Type of file
    
    Raises
    ------
    ValueError
        Raised if `file` is neither a request nor a results file.

    """
    if file.endswith('.res'):
        file_type = 'results'
        print('\n--------------\nResults file detected!\n--------------')
    elif file.endswith('.req'):
        file_type = 'request'
        print('\n--------------\nRequst file detected!\n--------------')
    else:
        raise AviewError('Files must be .res or .req files.')
    return file_type

def get_req_from_comment(comment, ans):
    """Gets the actual name of a request
    
    Parameters
    ----------
    comment : str
        Request comment
    results : Analysis
        Adams `:obj:Analysis` object
    
    Returns
    -------
    nt
        Adams results (`:obj:nt`)? object
        
    """
    for req_name in [r for r in ans.results.keys() if r.startswith('REQUEST_')]:
        if Adams.evaluate_exp(f'{req_name}.comment').lower() == comment.lower():
            return ans.results.get(req_name)

    raise AviewError(f'No request found with comment {comment}!')

class AviewError(Exception):
    """Use this error so ppt._get_log_errors will flag these errors.    
    
    """
    pass
            
if __name__ in ['aview_main', '__main__']:
    
    Adams.execute_cmd('def com echo=on')

    res_file = r'{{ res_file }}'
    reqs_to_get = {{ reqs_to_get }}
    t_min = {{ t_min }}
    t_max = {{ t_max }}
    output_file = r'{{ output_file }}'

    write_results(res_file, reqs_to_get, output_file, t_min, t_max)
    


"""This module is for sending scripts through Adams/PPT

"""
import os
import subprocess
import re
import time
import platform

import jinja2
from numpy import genfromtxt
from thornpy.signal import manually_clean_sig, remove_data_point, manually_clean_sigs, low_pass
from thornpy.signal import _clean_sig as clean_sig

from ..adamspy import get_log_errors, LOG_COMPLETE_PATTERN


LUNAR_SCRIPT_NAME = 'get_lunar_results.py'
GET_RESULTS_SCRIPT_NAME = 'get_results.py'
EDIT_RESULTS_SCRIPT_NAME = 'edit_results.py'
TEMP_OUTPUT_FILENAME = 'results.tmp'

LOG_NAME = 'aview.log'

TMPLT_ENV = jinja2.Environment(
    loader=jinja2.PackageLoader('adamspy.postprocess', 'aview_scripts'),
    autoescape=False,
    keep_trailing_newline=True,
    trim_blocks=True,
    lstrip_blocks=True
)

_TIMEOUT = 300

def get_results(res_file, reqs_to_get, t_min=None, t_max=None, _just_write_script=False, timeout=_TIMEOUT):
    """Gets results from an Adams results (.res) file.

    Example
    -------
    >>> result_file = 'example.res'
    >>> t_min = 70
    >>> t_max = 80
    >>> reqs_to_get = {}
    >>> reqs_to_get['MSE'] = ['Instantaneous_Bottom_MSE', 'Filtered_Surface_MSE']
    >>> reqs_to_get['ROP_controls'] = ['Command_ROP', 'True_WOB']
    >>> requests, units = get_results(result_file, reqs_to_get, t_min, t_max)

    Note
    ----
    This funciton only works with Requests.  It does not work with Result Sets.

    Note
    ----
    This function only works with xml results files.

    Parameters
    ----------
    result_file : str
        Filename of an Adams results (.res) file
    reqs_to_get : dict
        Dictionary of requests to extract (the default is None, which gets all results)
    t_min : float, optional
        Minimum time for which to extract results (the default is None)
    t_max : float, optional
        Maximum time for which to extract results (the default is None)

    Returns
    -------
    dict
        Dictionary of request data

    """
    template = TMPLT_ENV.from_string(open(os.path.join(os.path.dirname(__file__), 'aview_scripts', GET_RESULTS_SCRIPT_NAME)).read())
    working_directory = os.path.dirname(res_file)

    script_filename = _get_unique_filename(GET_RESULTS_SCRIPT_NAME)
    output_filename = _get_unique_filename(TEMP_OUTPUT_FILENAME)

    with open(os.path.join(working_directory, script_filename), 'w') as fid:
        fid.write(template.render({'res_file': os.path.split(res_file)[-1], 'reqs_to_get': reqs_to_get, 't_min': t_min, 't_max': t_max, 'output_file': output_filename}))

    if _just_write_script is False:
        # Delete the aview.log file
        try:
            os.remove(os.path.join(working_directory, LOG_NAME))
        except FileNotFoundError:
            pass
        except PermissionError:
            pass

        # Run the postprocessor
        if platform.system() == 'Windows':
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            subprocess.Popen('"{}" aview ru-s b {}'.format(os.environ['ADAMS_LAUNCH_COMMAND'], script_filename), cwd=working_directory, startupinfo=startupinfo)

        else:
            subprocess.Popen([os.environ['ADAMS_LAUNCH_COMMAND'], '-c', 'aview', 'ru-standard', 'b', script_filename, 'exit'], cwd=working_directory)

        # Wait for complete
        _wait(os.path.join(working_directory, LOG_NAME), timeout=timeout)

        # Check the log file for errors
        get_log_errors(os.path.join(working_directory, LOG_NAME))

        # Read and return the results
        data = genfromtxt(os.path.join(working_directory, output_filename), delimiter=',', names=True, dtype=None)

        os.remove(os.path.join(working_directory, script_filename))
        os.remove(os.path.join(working_directory, output_filename))

        output_dict = {'time': list(data['time'])}
        for res in reqs_to_get:
            output_dict[res] = {}
            for comp in reqs_to_get[res]:
                output_dict[res][comp] = list(data[f'{res}_{comp}'])

        return output_dict

def write_results(res_file, input_dict):
    # TODO
    return

def edit_results(res_file, input_dict, new_res_file=None, _just_write_script=False, timeout=_TIMEOUT):
    template = TMPLT_ENV.from_string(open(os.path.join(os.path.dirname(__file__), 'aview_scripts', EDIT_RESULTS_SCRIPT_NAME)).read())
    working_directory = os.path.dirname(res_file)

    new_res_file = os.path.split(res_file)[-1] if new_res_file is None else os.path.split(new_res_file)[-1]

    script_name = _get_unique_filename(EDIT_RESULTS_SCRIPT_NAME)

    with open(os.path.join(working_directory, script_name), 'w') as fid:
        fid.write(template.render({'res_file': os.path.split(res_file)[-1], 'reqs_to_edit': input_dict, 'output_file': new_res_file}))

    if _just_write_script is False:
        # Delete the aview.log file
        try:
            os.remove(os.path.join(working_directory, LOG_NAME))
        except FileNotFoundError:
            pass
        except PermissionError:
            pass

        if platform.system() == 'Windows':
            # Run the postprocessor
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            subprocess.Popen('"{}" aview ru-s b {}'.format(os.environ['ADAMS_LAUNCH_COMMAND'], script_name), cwd=working_directory, startupinfo=startupinfo)

        else:
            subprocess.Popen([os.environ['ADAMS_LAUNCH_COMMAND'], '-c', 'aview', 'ru-s', 'b', script_name, 'exit'], cwd=working_directory)

        # Wait for complete
        _wait(os.path.join(working_directory, LOG_NAME), timeout=timeout)

        # Check the log file for errors
        get_log_errors(os.path.join(working_directory, LOG_NAME))

def get_lunar_results(res_files, reqs_to_get, t_min, t_max, output_file, _just_write_script=False, timeout=_TIMEOUT):

    template = TMPLT_ENV.from_string(open(os.path.join(os.path.dirname(__file__), 'aview_scripts', LUNAR_SCRIPT_NAME)).read())

    working_directory = os.path.dirname(res_files[0])

    script_name = _get_unique_filename(LUNAR_SCRIPT_NAME)

    with open(os.path.join(working_directory, script_name), 'w') as fid:
        fid.write(template.render({'res_files': res_files, 'reqs_to_get': reqs_to_get, 't_min': t_min, 't_max': t_max, 'output_suffix': output_file}))

    if _just_write_script is False:
        # Delete the aview.log file
        try:
            os.remove(os.path.join(working_directory, LOG_NAME))
        except FileNotFoundError:
            pass

        # Run the postprocessor
        if platform.system() == 'Windows':
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            subprocess.Popen('"{}" aview ru-s b {}'.format(os.environ['ADAMS_LAUNCH_COMMAND'], script_name), cwd=working_directory, startupinfo=startupinfo)

        else:
            subprocess.Popen([os.environ['ADAMS_LAUNCH_COMMAND'], 'aview', 'ru-s', 'b', script_name], cwd=working_directory)

        # Wait for complete
        _wait(os.path.join(working_directory, LOG_NAME), timeout=timeout)

        os.remove(os.path.join(working_directory, script_name))

        # Make a list of the files that are written
        res_output_files = [os.path.splitext(output_file)[0] + '_time' + os.path.splitext(output_file)[-1]]
        for res_name, res_comps in reqs_to_get.items():
            for res_comp in res_comps:
                full_res_name = f'{res_name}_{res_comp}'
                res_output_files.append(os.path.splitext(output_file)[0] + '_' + full_res_name + os.path.splitext(output_file)[-1])

        return res_output_files

def _wait(log_file, sleep_time=0.2, timeout=300):
    """Waits for the log file to write the last line of the macro

    Parameters
    ----------
    log_file : str
        filename of log file
    sleep_time : float, optional
        Time between checks, by default 0.2
    timeout : int, optional
        During after which to time out, by default 300

    """

    for _i in range(int(timeout/sleep_time)):

        ppt_log_file_exists = os.path.exists(log_file)

        if ppt_log_file_exists is True:
            # If ppt.log exists, open it and see if the results have been loaded
            with open(log_file, 'r') as fid:
                text = fid.read()
            if re.search(LOG_COMPLETE_PATTERN, text):
                break

        time.sleep(sleep_time)

def manually_remove_spikes(res_file, reqs_to_clean, reqs_to_check=None, t_min=None, t_max=None, _just_write_script=False, timeout=_TIMEOUT, _inplace=False):
    """Allows the user to manually scan through the result sets to pick out points to eliminate.

    Parameters
    ----------
    res_file : str
        Adams Results (.res) filename
    reqs_to_clean : dict, optional
        Nested dictionary of result sets and result components to clean
    reqs_to_check : dict
        Nested dictionary of result sets and result components to check for spikes, by default same as reqs_to_clean
    t_min : float, optional
        Minumum simulation time to clean, by default None
    t_max : float, optional
        Maximum simulation time to clean, by default None
    timeout : float, optional
        Number of seconds to wait for results to load before timing out, by default _TIMEOUT

    Returns
    -------
    dict
        Nested dictionary of cleaned results

    """
    if reqs_to_check is None:
        reqs_to_check = reqs_to_clean

    results = get_results(res_file, reqs_to_clean, t_min=t_min, t_max=t_max, _just_write_script=_just_write_script, timeout=timeout)

    time_sig = results['time']

    # Remove the spikes
    for (res, res_comps) in [(r, rc) for r, rc in results.items() if r in reqs_to_check]:
        for (res_comp, values) in [(rc, v) for rc, v in res_comps.items() if rc in reqs_to_check[res]]:
            results[res][res_comp], i_mod = manually_clean_sig(time_sig, values, indices=True)

            # If a modification was made to the signal, make that modification to the rest of the signals
            if i_mod != []:

                # Loop over all the other results
                for (other_res, other_res_comps) in [(r, rc) for r, rc in results.items() if r != 'time']:
                    for (other_res_comp, other_values) in [(rc, v) for rc, v in other_res_comps.items() if not (other_res == res and rc == res_comp)]: #pylint: disable=no-member
                        for i in i_mod:
                            results[other_res][other_res_comp] = remove_data_point(time_sig, other_values, i)

    # Update the analysis files
    edit_results(res_file, results)

    # Return the cleaned results
    return results

def filter_results(res_file, reqs_to_clean, freq_cutoff, N_filter=5, reqs_to_check=None, t_min=None, t_max=None, _just_write_script=False, timeout=_TIMEOUT, _inplace=False, return_raw=False):
    """Similar to `manually_remove_spikes`, but allows user to plot the signals in batches.  Instead of passing a dictionary for the `reqs_to_check` argument, pass a list of dictionaries and the results in each dictionary in the list will be plotted together.

    Parameters
    ----------
    res_file : str
        Adams Results (.res) filename
    reqs_to_clean : dict
        Nested dictionary of result sets and result components to clean
    freq_cutoff : float
        Cutoff freqency of filter in Hz
    N_filter : int
        Order of filter
    reqs_to_check : list of dicts
        list of nested dictionary of result sets and result components to check for spikes, by default same as reqs_to_clean
    t_min : float, optional
        Minumum simulation time to clean, by default None
    t_max : float, optional
        Maximum simulation time to clean, by default None
    timeout : float, optional
        Number of seconds to wait for results to load before timing out, by default _TIMEOUT

    Returns
    -------
    dict
        Nested dictionary of cleaned results

    """
    if reqs_to_check is None:
        reqs_to_check = [reqs_to_clean]

    results = get_results(res_file, reqs_to_clean, t_min=t_min, t_max=t_max, _just_write_script=_just_write_script, timeout=timeout)
    time_sig = results.pop('time')

    filtered_results = {}
    for res_name, res_comps in results.items():
        filtered_results[res_name] = {}
        for res_comp, values in res_comps.items():                                          # pylint: disable=no-member

            cleaned_sig, _, _ = clean_sig(values, 3)

            filtered_results[res_name][res_comp], _ = low_pass(cleaned_sig, time_sig, freq_cutoff, N=N_filter)

    # Return the cleaned results
    if return_raw is True:
        return {'time': time_sig, **filtered_results}, {'time': time_sig, **results}
    else:
        return {'time': time_sig, **filtered_results}

def manually_remove_spikes_batch(res_file, reqs_to_clean, reqs_to_check=None, t_min=None, t_max=None, _just_write_script=False, timeout=_TIMEOUT, _inplace=False):
    """Similar to `manually_remove_spikes`, but allows user to plot the signals in batches.  Instead of passing a dictionary for the `reqs_to_check` argument, pass a list of dictionaries and the results in each dictionary in the list will be plotted together.

    Parameters
    ----------
    res_file : str
        Adams Results (.res) filename
    reqs_to_clean : dict
        Nested dictionary of result sets and result components to clean
    reqs_to_check : list of dicts
        list of nested dictionary of result sets and result components to check for spikes, by default same as reqs_to_clean
    t_min : float, optional
        Minumum simulation time to clean, by default None
    t_max : float, optional
        Maximum simulation time to clean, by default None
    timeout : float, optional
        Number of seconds to wait for results to load before timing out, by default _TIMEOUT

    Returns
    -------
    dict
        Nested dictionary of cleaned results

    """
    if reqs_to_check is None:
        reqs_to_check = [reqs_to_clean]

    results = get_results(res_file, reqs_to_clean, t_min=t_min, t_max=t_max, _just_write_script=_just_write_script, timeout=timeout)

    time_sig = results['time']

    # Create a flag indicating if the results have been modified and need to be rewritten
    results_modified = False

    for batch_to_check in reqs_to_check:

        # Make a list/batch of values to clean
        values_to_check = []
        for (res, res_comps) in [(r, rc) for r, rc in results.items() if r in batch_to_check]:
            for (res_comp, values) in [(rc, v) for rc, v in res_comps.items() if rc in batch_to_check[res]]:

                values_to_check.append(values)

        _, i_mod = manually_clean_sigs(time_sig, values_to_check, indices=True)

        # If a modification was made to the signal, make that modification to the rest of the signals
        if i_mod != []:

            # Loop over all the results
            for (res, res_comps) in [(r, rc) for r, rc in results.items() if r != 'time']:
                for (res_comp, values) in res_comps.items():
                    for i in i_mod:
                        results[res][res_comp] = remove_data_point(time_sig, values, i)

            # Flag that the results have been modified
            results_modified = True

    # If the results were modified, update the analysis files
    if results_modified is True:
        edit_results(res_file, results)

    # Return the cleaned results
    return results

def _get_unique_filename(filename):
    if os.path.exists(filename):

        for i in range(9999):
            new_name, ext = os.path.splitext(filename)
            new_name = new_name + f'_{i+1}'
            if not os.path.exists(new_name + ext):
                filename = new_name + ext
                break

    return filename



from contextlib import contextmanager
import os
from pathlib import Path
from platform import system
import re
import shutil
import subprocess
from tempfile import NamedTemporaryFile, TemporaryDirectory
import time
from typing import List

from ..adamspy import LOG_COMPLETE_PATTERN, get_log_errors

START_SCRIPT_NAMES: List[str] = ['aviewBS.cmd', 'aview.cmd', 'aviewAS.cmd']


def run_script(script_file: Path, cwd: Path = None, delete_log=True):
    """Runs the commands in `:arg:script_file` in an isolated Adams View session. Ignores any startup
    scripts (i.e. aviewBS.cmd, aview.cmd, aviewAS.cmd) in the working directory.

    Parameters
    ----------
    cmds : List[str]
        Commands to run (Adams View Command Language)
    cwd : Path
        Working directory
    """
    script_file = Path(script_file)
        
    if cwd is None:
        # If cwd not provided, set equal to script parent directory
        cwd = Path(script_file).parent
        delete_script = False

    elif Path(cwd) != Path(script_file).parent:
        # If cwd provided, and not equal to script parent directory, copy the script to a temporary
        # file in cwd
        with NamedTemporaryFile(mode='w', dir=cwd, suffix='.cmd', delete=False) as _fid:
            _fid.write(script_file.read_text())
            script_file = Path(_fid.name).resolve()
        
        # Flag to delete the script later
        delete_script = True
    
    else:
        # If cwd is provided **AND** it is equal to the script parent directory, just make sure
        # it's a Path object
        cwd = Path(cwd)
        delete_script = False

    # Get a unique log file name (NOTE: This also *creates* the log file in the process)
    _fid = NamedTemporaryFile(dir=cwd, suffix='.log', delete=False)
    _fid.close()
    log_file = Path(_fid.name)
    
    # Delete the file
    log_file.unlink()

    with no_start_scripts(cwd):

        # Change the log file name
        (cwd / 'aviewBS.cmd').write_text(f'file log_file file_name = "{log_file.name}"\n')

        # Run the commands in Adams
        if system() == 'Windows':
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            subprocess.Popen('"{}" aview ru-s b {}'.format(os.environ['ADAMS_LAUNCH_COMMAND'], script_file.name), cwd=cwd, startupinfo=startupinfo)

        else:
            subprocess.Popen([os.environ['ADAMS_LAUNCH_COMMAND'], '-c', 'aview', 'ru-standard', 'b', script_file.name, 'exit'], cwd=cwd)

        # Wait for complete
        _wait(log_file)

        # Remove temporary startup files
        for file in (f for f in (*START_SCRIPT_NAMES, 'aview.log') if (cwd / f).exists()):
            (cwd / file).unlink()

        if delete_script:
            script_file.unlink()

    log = Path(log_file).read_text()

    # Check the log file for errors
    try:
        get_log_errors(log_file)

    finally:

        # Delete the log
        if delete_log is True:
            log_file.unlink()

    return log


def run_commands(cmds: List[str], cwd: Path, delete_log=True):
    """Runs the commands in `:arg:cmds` in an isolated Adams View session. Ignores any startup
    scripts (i.e. aviewBS.cmd, aview.cmd, aviewAS.cmd) in the working directory.

    Parameters
    ----------
    cmds : List[str]
        Commands to run (Adams View Command Language)
    cwd : Path
        Working directory
    """
    # Write the commands
    with NamedTemporaryFile(dir=cwd, mode='w', suffix='.cmd', delete=False) as fid:
        fid.write('\n'.join(cmds))
        script_file = Path(fid.name)

    return run_script(script_file, cwd, delete_log=delete_log)


@contextmanager
def no_start_scripts(cwd: Path, skip: List[str] = None, preserve_log_file=True):
    """Moves all startup script files *NOT* named in `:arg:skip` to a temporary directory on entry.
    Moves them back on exit.

    Parameters
    ----------
    cwd : Path
        Current working directory
    scripts : str, optional
        Names of scripts to move. Nonexistent files are ignored, by default ['aviewBS.cmd', 'aview.cmd', 'aviewAS.cmd']
    """
    skip = skip or []
    cwd = Path(cwd)
    with TemporaryDirectory() as twd_:
        twd = Path(twd_).resolve()
        for script_name in [s for s in START_SCRIPT_NAMES if (cwd / s).is_file() and s not in skip]:
            shutil.move(cwd / script_name, twd / script_name)

        if preserve_log_file and (cwd / 'aview.log').is_file():
            shutil.move(cwd / 'aview.log', twd / 'aview.log')

        try:
            yield

        finally:
            for script_name in [s for s in START_SCRIPT_NAMES if (twd / s).is_file() and s not in skip]:
                if (cwd / script_name).is_file():
                    os.remove(cwd / script_name)
                shutil.move(twd / script_name, cwd / script_name)

            if preserve_log_file and (twd / 'aview.log').is_file():
                if (cwd / 'aview.log').is_file():
                    os.remove(cwd / 'aview.log')
                shutil.move(twd / 'aview.log', cwd / 'aview.log')


def _wait(log_file, sleep_time=0.2, timeout=300):
    """Waits for the log file to write the last line of the macro

    Parameters
    ----------
    log_file : str
        filename of log file
    sleep_time : float, optional
        Time between checks, by default 0.2
    timeout : int, optional
        During after which to time out, by default 300

    """

    for _i in range(int(timeout/sleep_time)):

        if Path(log_file).exists() is True:

            with open(log_file, 'r') as fid:
                text = fid.read()
            if re.search(LOG_COMPLETE_PATTERN, text):
                break

        time.sleep(sleep_time)
    time.sleep(sleep_time)



"""
adamspy is a set of python tools for manipulating MSC Adams files and data

Available subpackages
---------------------
adripy
    A set of python tools for manipulating MSC Adams Drill files and data

Available modules
-----------------
adamspy
    Functions for manipulating adams files and data
"""
from . import adripy
from . import waterfall
name = 'adamspy'





import os
import glob
import shutil
import unittest

import matplotlib.pyplot as plt
import pandas as pd

from adamspy.postprocess.ppt import manually_remove_spikes, filter_results, get_results

TEST_ANALYSIS_DIR = os.path.join('test', 'files', 'real_analysis_with_spikes')

LOWER_TORQUE_MEAS_NAME = 'clutch_torque_lower'
UPPER_TORQUE_MEAS_NAME = 'clutch_torque_upper'
UPPER_SPRING_STRESS_MEAS = {f'spring_stress_U{el+2}': ['Q'] for el in range(30)}
LOWER_SPRING_STRESS_MEAS = {f'spring_stress_L{el+2}': ['Q'] for el in range(28)}
SPRING_STRESS_MEAS = {**LOWER_SPRING_STRESS_MEAS, **UPPER_SPRING_STRESS_MEAS}
MEASURES = {**SPRING_STRESS_MEAS, LOWER_TORQUE_MEAS_NAME: ['Q'], UPPER_TORQUE_MEAS_NAME: ['Q']}

class Test_FilterSpikes(unittest.TestCase):

    def setUp(self):
        self.temp_ans_dir = TEST_ANALYSIS_DIR + '_tmp'        
        self._setup_files()
        self.res_to_edit = glob.glob(os.path.join(self.temp_ans_dir, '*.res'))[0]

    def test_filter_spikes(self):
        filtered_results, raw_results = filter_results(self.res_to_edit, MEASURES, 20, return_raw=True)       
        df_filt = convert_dict_to_df(filtered_results).set_index('time', drop=True)
        df_raw = convert_dict_to_df(raw_results).set_index('time', drop=True)

        for meas in SPRING_STRESS_MEAS.keys():
            _fig, ax = plt.subplots()
            ax.plot(df_filt[meas]/1000, label='filtered')
            ax.plot(df_raw[meas]/1000, '.', markersize=5, alpha=.8, label='raw')
            ax.set_xlim([4, 5.5])
            ax.grid()
            ax.set_ylabel('Stress (ksi)')
            ax.set_xlabel('Time (sec)')
            ax.legend()
        
        for meas in [LOWER_TORQUE_MEAS_NAME, UPPER_TORQUE_MEAS_NAME]:
            _fig, ax = plt.subplots()
            ax.plot(df_filt[meas], label='filtered')      
            ax.plot(df_raw[meas], '.', markersize=5, alpha=.8, label='raw')               
            ax.set_xlim([4, 5.5])         
            ax.grid()
            ax.set_xlabel('Time (sec)')
            ax.set_ylabel('Torque (lbf-in)')
            ax.legend()

        plt.show()
        self.assertTrue(False)

    def tearDown(self):
        """Deletes the temporary working directory.

        """
        shutil.rmtree(self.temp_ans_dir)
            
    def _setup_files(self):
        """Copies the test analysis files to a temporary working directory

        """
        if os.path.exists(self.temp_ans_dir) is True:
            shutil.rmtree(self.temp_ans_dir)

        os.makedirs(self.temp_ans_dir)
        
        for file in glob.glob(os.path.join(TEST_ANALYSIS_DIR, '*')):
            shutil.copyfile(file, os.path.join(self.temp_ans_dir, os.path.split(file)[-1]))   



def convert_dict_to_df(data):
    reform = {'time': data['time'], **{f'{outer_key}': values for outer_key, inner_dict in data.items() if outer_key != 'time' for inner_key, values in inner_dict.items()}}
    return pd.DataFrame(reform)


import os

import pandas as pd

from ..postprocess.ppt import get_results

RESULT_SETS = {
    'tire_forces': ['normal', 'longitudinal'],
    'tire_kinematics': ['longitudinal_slip', 'slip_angle']
}

SOFT_SOIL_RESULTS_SETS = {
    'tire_soft_soil_main': ['tire_deflection', 'loaded_radius', 'rolling_radius', 'substitution_radius', 'total_sinkage', 'rotational_velocity'],
    'tire_soft_soil_stress': ['entry_angle', 'exit_angle', 'max_stress_angle', 'max_normal_stress'],
    'tire_soft_soil_deform': ['z_elastic']
}

RESULTS_MAP = {
    'F_x': ['tire_forces', 'longitudinal', 0.224809],
    'F_n': ['tire_forces', 'normal', 0.224809],
    'k': ['tire_kinematics', 'longitudinal_slip', 1],
    'alpha': ['tire_kinematics', 'slip_angle', 1]
}

SOFT_SOIL_RESULTS_MAP = {    
    'f_0': ['tire_soft_soil_main', 'tire_deflection', 0.0393701],
    'h': ['tire_soft_soil_main', 'total_sinkage', 0.0393701],
    'h_e': ['tire_soft_soil_deform', 'z_elastic', 0.0393701],
    'theta_f': ['tire_soft_soil_stress', 'entry_angle', 1],
    'theta_r': ['tire_soft_soil_stress', 'exit_angle', 1],
    'theta_m': ['tire_soft_soil_stress', 'max_stress_angle', 1],
    'R_s': ['tire_soft_soil_main', 'substitution_radius', 0.0393701],
    'R_l': ['tire_soft_soil_main', 'loaded_radius', 0.0393701],
    'R_r': ['tire_soft_soil_main', 'rolling_radius', 0.0393701],
    'w': ['tire_soft_soil_main', 'rotational_velocity', 1],
    'sig_max': ['tire_soft_soil_stress', 'max_normal_stress', 145.038]
}


def map_results(res_dict, soft_soil=False):
    df = pd.DataFrame({'t': res_dict['time']}) 

    results_map = RESULTS_MAP if soft_soil is False else {**RESULTS_MAP, **SOFT_SOIL_RESULTS_MAP}

    for name, res in results_map.items():
        df[name] = res_dict[res[0]][res[1]]
        df[name] = df[name]*res[2]

    return df

def get_tire_testrig_results(res_file, soft_soil=False, res_file_units='mmks'):    
    result_sets = RESULT_SETS if soft_soil is False else {**RESULT_SETS, **SOFT_SOIL_RESULTS_SETS}
    res_dict = get_results(res_file, result_sets)
    return map_results(res_dict, soft_soil=soft_soil)


"""Functions for manipulating adams files and data
"""
import os
import re
import subprocess
import platform

XMT_PATTERN = re.compile('\\s*file_name\\s*=\\s*"?.+\\.xmt_txt"?\\s*')
LOG_FILE_ERROR_PATTERN = '! \\S*Error: '
LOG_COMPLETE_PATTERN = '! Command file is exhausted, batch run is finished.'


def get_simdur_from_msg(msg_file):
    """Reads an Adams message file (.msg) and returns the total duration of the simulation.

    Parameters
    ----------
    msg_file : str
        Filepath to an Adams message file (.msg)

    Raises
    ------
    RuntimeError
        Returned if no simulation end time was found in the specified message file

    Returns
    -------
    float
        Total duration of the simulation

    """
    found = False
    with open(msg_file, 'r') as fid:
        for line in fid:
            if re.match(' *command: sim(ulate)?/dyn(anmic)?.*, *end *=.*', line.lower()):
                duration = float(re.split('end *=', line.lower().replace(' ', ''))[-1]
                                 .split(',')[0])
                found = True

    # Raise an error if no duration found
    if not found:
        raise RuntimeError('No simulation end time was found in the specified message file!')

    return duration


def get_simdur_from_acf(acf_file):
    """Reads an Adams command file (.acf) and returns the total duration of the simulation.

    Parameters
    ----------
    acf_file : str
        Filepath to an Adams message file (.msg)

    Raises
    ------
    RuntimeError
        Returned if no simulation end time was found in the specified message file

    Returns
    -------
    float
        Total duration of the simulation

    """
    found = False
    with open(acf_file, 'r') as fid:
        for line in fid:
            if re.match('sim(ulate)?/dyn(anmic)?.*, *end *=.*', line.lower()):
                duration = float(re.split('end *=', line.lower().replace(' ', ''))[-1]
                                 .split(',')[0])
                found = True

    # Raise an error if no duration found
    if not found:
        raise RuntimeError('No simulation end time was found in the specified acf file!')

    return duration


def modify_xmt_path(cmd_file, new_xmt_path):
    with open(cmd_file, 'r') as fid_in, open(cmd_file.replace('.cmd', '.tmp'), 'w') as fid_out:
        for line in fid_in:
            if XMT_PATTERN.match(line) is None:
                fid_out.write(line)
            else:
                line_parts = line.split('"')
                fid_out.write('"'.join([line_parts[0], new_xmt_path, line_parts[-1]]))
                original_path = line_parts[1]
    os.remove(cmd_file)
    os.rename(cmd_file.replace('.cmd', '.tmp'), cmd_file)
    return original_path


def get_n_threads(adm_file):
    """Searches `adm_file` for the NTHREADS statement and returns its value.

    Parameters
    ----------
    adm_file : str
        Path to an Adams Dataset (.adm) file

    Returns
    -------
    int
        Number of threads set `adm_file`

    """
    n_threads = 1
    with open(adm_file, 'r') as fid:
        for line in fid:

            # If at the NTHREADS statement, rewrite it
            if re.match('^,[ \\t]*nthreads[ \\t]*=[ \\t]*\\d$', line, flags=re.I):
                n_threads = int(line.split('=')[1].strip())

    return n_threads


def set_n_threads(adm_file, n_threads):
    """Changes or creates the NTHREADS option on the PREFERENCES statement in `adm_file`.

    Parameters
    ----------
    adm_file : str
        File path to an Adams Dataset (.adm) file
    n_threads : int
        Number of threads to use when running the model specified in `adm_file`

    """
    found = False
    with open(adm_file, 'r') as fid_old, open(adm_file + '.tmp', 'w') as fid_new:
        for line in fid_old:

            # If at the NTHREADS statement, rewrite it
            if re.match('^,[ \\t]*nthreads[ \\t]*=[ \\t]*\\d$', line, flags=re.I):
                fid_new.write(f', NTHREADS = {n_threads}\n')
                found = True

            # If the end is reached and the NTHREADS statement isn't found, create it
            elif re.match('^end[ \\t]*$', line, re.I) and not found:
                fid_new.write(f'PREFERENCES/\n, NTHREADS = {n_threads}\n!\n')
                fid_new.write(line)

            # If at a normal line, write it
            else:
                fid_new.write(line)

    # Delete the old adm file and replace with modified
    os.remove(adm_file)
    os.rename(adm_file + '.tmp', adm_file)


def solve(acf_file, wait=False, use_adams_car=False):
    """Runs Adams Solver to solve the model specified in `acf_file`

    Parameters
    ----------
    acf_file : str
        Path to an Adams Command (.acf) File

    """
    file = os.path.split(acf_file)[-1]
    cwd = os.path.dirname(acf_file) if os.path.dirname(acf_file) != '' else os.getcwd()

    if platform.system() == 'Windows':
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW

        if use_adams_car is False:
            command = '"{}" ru-s "{}"'.format(os.environ['ADAMS_LAUNCH_COMMAND'], file)
        else:
            command = '"{}" acar ru-solver "{}"'.format(os.environ['ADAMS_LAUNCH_COMMAND'], file)

        proc = subprocess.Popen(command, cwd=cwd, startupinfo=startupinfo)

    else:
        if use_adams_car is False:
            command = [os.environ['ADAMS_LAUNCH_COMMAND'], '-c', 'ru-standard', 'i', file, 'exit']
        else:
            command = [os.environ['ADAMS_LAUNCH_COMMAND'], '-c', 'acar', 'ru-solver', 'i', file, 'exit']

        proc = subprocess.Popen(command, cwd=cwd)

    if wait:
        proc.wait()

    return proc


def get_log_errors(log_file):
    """Checks the log file for errors of the type AviewError.
    
    Parameters
    ----------
    log_file : str
        Filename of aview log file (usulally aview.log)

    """
    with open(log_file, 'r') as fid:
        lines = fid.readlines()

    errors = []
    for line in lines:
        if re.search(LOG_FILE_ERROR_PATTERN, line, flags=re.IGNORECASE):
            errors.append(line[2:])

    if errors:
        raise AviewError(''.join(errors))


class AdmFileError(Exception):
    pass


class AviewError(Exception):
    """Raise this error to if a known error occurs in the log file.
    
    """
    pass